@@ -6,8 +6,7 @@ from .._summary import summary
 
 from .._training import Trainer
 from ..layers import KerasLayer
-from ..metrics import Metric
-from ..metrics import _create_metric
+from ..metrics import _to_metrics_dic
 from ..losses import _create_loss
 from ..optimizers import _create_optimizer
 from ..activations import _create_activation
@@ -37,8 +36,8 @@ class Model(torch.nn.Module):
             return False
         self._has_keras_layer = dfs(self)
 
-    def forward(self, x):
-        return self.model(x)
+    def forward(self, *args):
+        return self.model(*args)
 
     def count_params(self) -> int:
         """Count the total number of scalars composing the weights."""
@@ -48,17 +47,25 @@ class Model(torch.nn.Module):
         """Return all trainable parameters of the model."""
         return filter(lambda p: p.requires_grad, self.parameters())
 
-    ########## keras-style methods below ##########
-
     @torch.no_grad()
-    def build(self, input_shape, dtype=torch.float32):
+    def build(self, input_shape, dtype=torch.float32, batch_size=8):
         """Build the model when it contains `KerasLayer`."""
-        if self._has_keras_layer:
-            batch_shape = [2] + list(input_shape)
-            probe_input = torch.zeros(size=batch_shape).to(dtype=dtype)
-            self.model(probe_input)
+        if isinstance(input_shape[0], int):
+            input_shape = [input_shape]
+
+        if not isinstance(dtype, list):
+            dtype = [dtype] * len(input_shape)
+
+        assert len(dtype) == len(input_shape)
+
+        device = self.trainer.device if self.compiled else 'cpu'
+
+        batch_shapes = [ [batch_size]+list(i) for i in input_shape]
+        probe_inputs = [torch.zeros(size=sz).to(dtype=dt, device=device) for sz,dt in zip(batch_shapes, dtype)]
+        self.model(*probe_inputs)
+
+        self._probe_inputs = probe_inputs
         self.built = True
-        self.input_shape = list(input_shape)
         return self
 
     def _check_keras_layer(self):
@@ -67,19 +74,18 @@ class Model(torch.nn.Module):
                 "You need to build the model via `.build()` before this operation. Because it contains `KerasLayer`."
                 )
 
-    def summary(self, input_shape=None, depth=3, device=None):
+    def summary(self, depth=3, device=None):
         """Print a string summary of the network."""
         self._check_keras_layer()
-        if input_shape is None:
-            input_shape = self.input_shape
-        assert (input_shape is not None)
+        if not self.built:
+            raise AssertionError('Build the model first before you call `.summary()`.')
 
         if device is None:
             device = self.trainer.device if self.compiled else 'cpu'
+        
+        summary(self.model, self._probe_inputs, depth=depth, verbose=1, device=device)
 
-        summary(self.model, input_shape, depth=depth, verbose=1, device=device)
-
-    def compile(self, optimizer, loss, metrics=None, device=None):
+    def compile(self, optimizer, loss, metrics=None, epoch_metrics=None, device=None):
         """
         Configure the model for training.
 
@@ -92,6 +98,8 @@ class Model(torch.nn.Module):
         * `metrics`: List of metrics to be evaluated by the model during training. You can also use dict to specify the 
         abbreviation of each metric.
 
+        * `epoch_metrics`: List of non-linear metrics(e.g. ROC_AUC) that need to be evaluated on epoch end.
+
         * `device`: Device of the model and its trainer, if `None` 'cuda' will be used when `torch.cuda.is_available()` otherwise 'cpu'.
         """
         self._check_keras_layer()
@@ -101,40 +109,36 @@ class Model(torch.nn.Module):
         loss = _create_loss(loss)
         optimizer = _create_optimizer(optimizer, self.trainable_params())
 
-        m = OrderedDict({'loss': loss})
-        if isinstance(metrics, dict):
-            m.update(metrics)
-        elif isinstance(metrics, list):
-            for tmp_m in metrics:
-                tmp_m = _create_metric(tmp_m)
-                if isinstance(tmp_m, Metric):
-                    m[tmp_m.get_abbr()] = tmp_m
-                elif hasattr(tmp_m, '__call__'):
-                    m[tmp_m.__name__] = tmp_m
-                else:
-                    raise TypeError('Unsupported type.')
-        elif not (metrics is None):
-            raise TypeError('Argument `metrics` should be either a dict or list.')
+        batch_metrics = OrderedDict({'loss': loss})
+        batch_metrics.update(_to_metrics_dic(metrics))
+
+        epoch_metrics = _to_metrics_dic(epoch_metrics)
 
         self.to(device=device)
-        self.trainer = Trainer(model=self, optimizer=optimizer, loss=loss, metrics=m, device=device)
+        self.trainer = Trainer(model=self, optimizer=optimizer, loss=loss, metrics=batch_metrics, epoch_metrics=epoch_metrics, device=device)
         self.compiled = True
 
+    @property
+    def trainer_loop_configs(self):
+        assert self.compiled
+        return self.trainer._batch_training_loop, self.trainer._batch_validation_loop
 
     def fit_dl(self, train_loader, epochs,
                 val_loader=None,
                 callbacks=[],
                 verbose=1,
-                use_amp=False):
+                use_amp=False,
+                accum_grad_steps=1):
 
         assert self.compiled
         self.trainer.register_callbacks(callbacks)
-        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, use_amp=use_amp)
+        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, use_amp=use_amp, accum_grad_steps=accum_grad_steps)
 
         return history
 
 
     def fit(self, x, y, epochs, batch_size=32,
+                validation_batch_size=None,
                 validation_split=None, shuffle_val_split=True,
                 validation_data=None,
                 callbacks=[],
@@ -142,14 +146,15 @@ class Model(torch.nn.Module):
                 shuffle=True,
                 sample_weight=None,
                 num_workers=0,
-                use_amp=False
+                use_amp=False,
+                accum_grad_steps=1
                 ):
         """
         Train the model for a fixed number of epochs (iterations on a dataset).
 
         Args:
 
-        * `x` (`ndarray` or `torch.Tensor`): Input data 
+        * `x` (`ndarray` or `torch.Tensor` or list): Input data
 
         * `y` (`ndarray` or `torch.Tensor`): Target data
 
@@ -157,6 +162,8 @@ class Model(torch.nn.Module):
 
         * `batch_size` (int, default=32): Number of samples per gradient update
 
+        * `validation_batch_size` (int, default=None): Number of samples for each step on validation loop, if `None` will use `batch_size`
+
         * `validation_split` (float between 0 and 1): Fraction of the training data to be used as validation data
 
         * `shuffle_val_split` (bool, default=True): Whether to do shuffling when `validation_split` is provided
@@ -165,7 +172,7 @@ class Model(torch.nn.Module):
         
         * `callbacks` (list of `keras4torch.callbacks.Callback`): List of callbacks to apply during training
 
-        * `verbose` (int, default=1): 0, 1, or 2. Verbosity mode. 0 = silent, 1 = normal, 2 = compat
+        * `verbose` (int, default=1): 0, 1, or 2. Verbosity mode. 0 = silent, 1 = normal, 2 = brief
 
         * `shuffle` (bool, default=True): Whether to shuffle the training data before each epoch
 
@@ -177,16 +184,20 @@ class Model(torch.nn.Module):
         """
 
         assert self.compiled
+        if not isinstance(x, list) and not isinstance(x, tuple):
+            x = [x]
         x, y = to_tensor(x, y)
 
         assert not (validation_data != None and validation_split != None)
         has_val = validation_data != None or validation_split != None
         
-        train_set = TensorDataset(x, y)
+        train_set = TensorDataset(*x, y)
 
         if validation_data != None:
             x_val, y_val = to_tensor(validation_data[0], validation_data[1])
-            val_set = TensorDataset(x_val, y_val)
+            if not isinstance(x_val, list) and not isinstance(x_val, tuple):
+                x_val = [x_val]
+            val_set = TensorDataset(*x_val, y_val)
 
         if validation_split != None:
             val_length = int(len(train_set) * validation_split)
@@ -196,7 +207,7 @@ class Model(torch.nn.Module):
                 np.random.shuffle(idx)
             train_set, val_set = Subset(train_set, idx[:-val_length]), Subset(train_set, idx[-val_length:])
             if sample_weight is not None:
-                assert len(sample_weight) == len(x)
+                assert len(sample_weight) == len(y)
                 sample_weight = [sample_weight[i] for i in idx[:-val_length]]
 
         if sample_weight is not None:
@@ -205,26 +216,32 @@ class Model(torch.nn.Module):
         else:
             sampler = None
         
-        dl_kwargs = {'batch_size': batch_size, 'num_workers': _get_num_workers(num_workers)}
-        train_loader = DataLoader(train_set, shuffle=shuffle, sampler=sampler, **dl_kwargs)
-        val_loader = DataLoader(val_set, shuffle=False, **dl_kwargs) if has_val else None
+        num_workers = _get_num_workers(num_workers)
 
-        return self.fit_dl(train_loader, epochs, val_loader, callbacks, verbose, use_amp)
+        if validation_batch_size is None:
+            validation_batch_size = batch_size
+
+        train_loader = DataLoader(train_set, shuffle=shuffle, sampler=sampler, batch_size=batch_size, num_workers=num_workers)
+        val_loader = DataLoader(val_set, shuffle=False, batch_size=validation_batch_size, num_workers=num_workers) if has_val else None
+
+        return self.fit_dl(train_loader, epochs, val_loader, callbacks, verbose, use_amp, accum_grad_steps)
 
     @torch.no_grad()
     def evaluate(self, x, y, batch_size=32, num_workers=0, use_amp=False):
         """Return the loss value & metrics values for the model in test mode.\n\n    Computation is done in batches."""
+        if not isinstance(x, list) and not isinstance(x, tuple):
+            x = [x]
         x, y = to_tensor(x, y)
-        val_loader = DataLoader(TensorDataset(x, y), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
+        val_loader = DataLoader(TensorDataset(*x, y), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
         return self.evaluate_dl(val_loader, use_amp)
 
     @torch.no_grad()
     def evaluate_dl(self, data_loader, use_amp=False):
         assert self.compiled
-        return self.trainer.evaluate(data_loader, use_amp)
+        return dict(self.trainer.valid_on_epoch(data_loader, use_amp))
 
     @torch.no_grad()
-    def predict_dl(self, data_loader, device=None, output_numpy=True, activation=None):
+    def predict_dl(self, data_loader, device=None, output_numpy=True, activation=None, use_amp=False):
         self._check_keras_layer()
         if device == None:
             if self.compiled:
@@ -233,7 +250,17 @@ class Model(torch.nn.Module):
                 device = 'cuda' if torch.cuda.is_available() else 'cpu'
         self.eval().to(device=device)
 
-        outputs = [self(x_batch[0].to(device=device)) for x_batch in data_loader]
+        outputs = []
+        for batch in data_loader:
+            if not isinstance(batch, list):
+                batch = [batch]
+                
+            for i in range(len(batch)):
+                batch[i] = batch[i].to(device=device)
+
+            with torch.cuda.amp.autocast(enabled=use_amp):
+                outputs.append(self(*batch))
+
         outputs = torch.cat(outputs, dim=0)
 
         activation = _create_activation(activation)
@@ -243,13 +270,12 @@ class Model(torch.nn.Module):
         return outputs.cpu().numpy() if output_numpy else outputs
 
     @torch.no_grad()
-    def predict(self, inputs, batch_size=32, device=None, output_numpy=True, activation=None, num_workers=0):
-        """
-        Generate output predictions for the input samples.\n\n    Computation is done in batches.
-        """
-        inputs = to_tensor(inputs)
-        data_loader = DataLoader(TensorDataset(inputs), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
-        return self.predict_dl(data_loader, device=device, output_numpy=output_numpy, activation=activation)
+    def predict(self, x, batch_size=32, device=None, output_numpy=True, activation=None, num_workers=0, use_amp=False):
+        """Generate output predictions for the input samples.\n\n    Computation is done in batches."""
+        if not isinstance(x, list) and not isinstance(x, tuple):
+            x = [x]
+        data_loader = DataLoader(TensorDataset(*to_tensor(x)), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
+        return self.predict_dl(data_loader, device=device, output_numpy=output_numpy, activation=activation, use_amp=use_amp)
 
     def save_weights(self, filepath):
         """Equal to `torch.save(model.state_dict(), filepath)`."""
@@ -257,4 +283,4 @@ class Model(torch.nn.Module):
 
     def load_weights(self, filepath):
         """Equal to `model.load_state_dict(torch.load(filepath))`."""
-        self.load_state_dict(torch.load(filepath))
+        self.load_state_dict(torch.load(filepath))
\ No newline at end of file
